import json
import re
import sys
import rich
from rich.console import Console
from rich.json import JSON
from rich.markdown import Markdown
from rich.table import Table
from rich.rule import Rule

from chatblade import utils


console = Console()


def warn(msg):
    rich.print(f"[red]{msg}[/red]", file=sys.stderr)


def print_tokens(messages, token_stats, args):
    if args.only:
      args.roles = ["assistant"]
    else:
      args.roles = ["user", "assistant", "system"]
    print_messages(messages, args)
    console.print()
    table = Table(title="tokens/costs")
    table.add_column("Model", no_wrap=True)
    table.add_column("Tokens", no_wrap=True)
    table.add_column("Price", style="bold", justify="right")
    for token_stat in token_stats:
        table.add_row(
            token_stat.name,
            "{:d}".format(token_stat.tokens),
            "${:.6f}".format(token_stat.cost),
        )
    console.print(table)
    console.print(
        "[red] * estimated costs do not include the tokens that may be returned[/red]"
    )


def print_messages(messages, args):
    if "roles" not in args:
      if args.only:
        args.roles = ["assistant"]
      else:
        args.roles = ["user", "assistant"]
    if args.extract:
        extract_messages(messages, args)
    else:
        for message in messages:
            if message.role in args.roles:
                print_message(message, args)


COLORS = {"user": "blue", "assistant": "green", "system": "red"}


def print_message(message, args):
    printable = message.content
    if not args.raw:
        printable = detect_and_format_message(
            message.content, cutoff=1000 if message.role == "user" else None, theme=args.theme
        )
    if not args.no_format:
      console.print(Rule(message.role, style=COLORS[message.role]))

    if args.raw:
        print(message.content)
    else:
        console.print(printable)

    if not args.no_format:
      console.print(Rule(style=COLORS[message.role]))


def extract_messages(messages, args):
    message = messages[-1]
    if contains_json(message.content):
        print(extract_json(message.content))
    elif contains_block(message.content):
        print(extract_block(message.content))
    else:
        print(message.content.strip())


def detect_and_format_message(msg, cutoff=None, theme=None):
    if cutoff and len(msg) > cutoff:
        msg = "... **text shortened** ... " + msg[-cutoff:]
        return msg
    elif contains_json(msg):
        utils.debug(detected="json")
        return JSON(extract_json(msg))
    elif looks_like_markdown(msg):
        utils.debug(detected="markdown")
        theme = "monokai" if theme is None else theme
        return Markdown(msg,code_theme=theme)
    else:
        utils.debug(detected="regular")
        return msg


def extract_json_lists(str_lists, flatten=False):
    lists = [json.loads(extract_json(x)) for x in str_lists if contains_json(x)]
    if flatten:
        return json.dumps([item for sublist in lists for item in sublist])
    else:
        return json.dumps(lists)


def contains_block(str):
    if extract_block(str):
        return True
    return False


def extract_block(str):
    matches = re.findall(r"```[\w]*(.*?)```", str, re.DOTALL)
    try:
        return sorted(matches, key=lambda x: len(x))[-1].strip()
    except IndexError:
        return None


def looks_like_markdown(str):
    """very rudimentary, but avoids making things markdown that shouldn't be"""
    md_links = len(re.findall(r"\[[^]]+\]\(https?:\/\/\S+\)", str))
    md_text = len(re.findall(r"\s(__|\*\*)(?!\s)(.(?!\1))+(?!\s(?=\1))", str))
    md_blocks = len(re.findall(r"```(.*?)```", str, re.DOTALL))
    md_inline_blocks = len(re.findall(r"`[^`]+`", str)) - md_blocks

    md_blocks *= 2
    utils.debug(
        title="counted",
        md_links=md_links,
        md_text=md_text,
        md_inline_blocks=md_inline_blocks,
        md_blocks=md_blocks,
    )
    score = md_links + md_text + md_inline_blocks + md_blocks
    return score >= 2


def contains_json(str):
    try:
        extract_json(str)
    except ValueError:
        return False
    return True


def extract_json(str):
    """
    try to extract json from a string that may contain other lines before the json
    returns the json as a string or raises a ValueError if no json is found
    """
    lines_with_idxs = enumerate(str.splitlines())
    for idx, line in lines_with_idxs:
        if line.strip().startswith("{") or line.strip().startswith("["):
            return json.dumps(json.loads(" ".join(str.splitlines()[idx:])))

    raise ValueError("No json in string")



import glob
import re
import os

from . import storage


def list_sessions():
    """List names of sessions"""
    sess_paths = glob.glob(os.path.join(storage.get_cache_path(), "*.yaml"))
    return sorted(
        [
            re.sub("\\.yaml\\Z", "", os.path.basename(sess_path))
            for sess_path in sess_paths
        ]
    )


def rename_session(session, newname):
    """renames session
    Returns None on success, error string otherwise"""
    session_path = storage.get_session_path(session, True)
    if not session_path:
        return f"session {session} does not exist"
    new_session_path = storage.get_session_path(newname, True)
    if new_session_path:
        return f"session {newname} already exists"
    new_session_path = storage.get_session_path(newname)
    os.rename(session_path, new_session_path)


def delete_session(session):
    """deletes a session
    Returns None on success, error string otherwise"""
    session_path = storage.get_session_path(session, True)
    if not session_path:
        return f"session {session} does not exist"
    os.unlink(session_path)



import collections
import os
import yaml

import tiktoken
import openai
from openai._exceptions import OpenAIError

from . import utils, errors


class Message(collections.namedtuple("Message", ["role", "content"])):
    @staticmethod
    def represent_for_yaml(dumper, msg):
        val = []
        md = msg._asdict()

        for fie in msg._fields:
            val.append([dumper.represent_data(e) for e in (fie, md[fie])])

        return yaml.nodes.MappingNode("tag:yaml.org,2002:map", val)

    @classmethod
    def import_yaml(cls, seq):
        """instantiate from YAML provided representation"""
        return cls(**seq)


yaml.add_representer(Message, Message.represent_for_yaml)


CostConfig = collections.namedtuple("CostConfig", "name prompt_cost completion_cost")
CostCalculation = collections.namedtuple("CostCalculation", "name tokens cost")

costs = [
    CostConfig("gpt-3.5-turbo", 0.5, 1.5),
    CostConfig("gpt-4", 5, 15),
    CostConfig("gpt-4-1106-preview", 5, 15),
    CostConfig("gpt-4o-mini", 0.15, 0.6),
]


def get_tokens_and_costs(messages):
    return [
        CostCalculation(
            cost_config.name, *num_tokens_in_messages(messages, cost_config)
        )
        for cost_config in costs
    ]


def num_tokens_in_messages(messages, cost_config):
    """Returns the number of tokens used by a list of messages."""
    try:
        encoding = tiktoken.encoding_for_model(f"{cost_config.name}-0301")
    except KeyError:
        encoding = tiktoken.get_encoding("cl100k_base")
    num_tokens = 0
    cost = 0
    for i, message in enumerate(messages):
        msg_tokens = (
            4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
        )
        msg_tokens += len(encoding.encode(message.role))
        msg_tokens += len(encoding.encode(message.content))
        if i == len(messages) - 1 and message.role == "assistant":
            cost += cost_config.completion_cost * msg_tokens
        else:
            cost += cost_config.prompt_cost * msg_tokens
        num_tokens += msg_tokens
    if messages[-1].role == "user":
        num_tokens += 2  # every reply is primed with <im_start>assistant
        cost += cost_config.prompt_cost * 2
    return num_tokens, cost / 1000000


def init_conversation(user_msg, system_msg=None):
    system = [Message("system", system_msg)] if system_msg else []
    return system + [Message("user", user_msg)]


DEFAULT_OPENAI_SETTINGS = {
    "model": "gpt-3.5-turbo",
    "temperature": 0.1,
    "n": 1,
    "stream": False,
}


def map_from_stream(openai_gen):
    """maps a openai streaming generator a stream of Message with the
    final one being the completed Message"""
    role, message = None, ""
    for update in openai_gen:
        delta = [choice.delta for choice in update.choices][0]
        if delta.role:
            role = delta.role
        elif delta.content:
            message += delta.content
        yield Message(role, message)


def map_single(result):
    """maps a result to a Message"""
    response_message = [choice.message for choice in result.choices][0]
    return Message(response_message.role, response_message.content)


def build_client(config):
    if "OPENAI_API_AZURE_ENGINE" in os.environ:
        return openai.AzureOpenAI(
            api_key=config["openai_api_key"],
            azure_deployment=os.environ.get("OPENAI_API_AZURE_ENGINE"),
        )
    else:
        return openai.OpenAI(api_key=config["openai_api_key"], base_url=config["openai_base_url"])


def query_chat_gpt(messages, config):
    """Queries the chat GPT API with the given messages and config."""
    client = build_client(config)
    config = utils.merge_dicts(DEFAULT_OPENAI_SETTINGS, config)
    dict_messages = [msg._asdict() for msg in messages]
    try:
        result = client.chat.completions.create(messages=dict_messages, **config)
        if isinstance(result, openai._streaming.Stream):
            return map_from_stream(result)
        elif isinstance(result, openai.types.chat.ChatCompletion):
            return map_single(result)
        else:
            raise ValueError(f"unexpected result openai: {result}")
    except OpenAIError as e:
        raise errors.ChatbladeError(f"openai error: {e}")



from rich.pretty import pprint

CONSOLE_DEBUG_LOGGING = False

scratch_session = "last"


class DotDict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def merge_dicts(dict1, dict2):
    """merge 2 dicts with priority from dict2, but only for keys known by dict1"""
    merged = {**dict1, **dict2}
    merged = {k: v for k, v in merged.items() if v is not None and k in dict1}
    return DotDict(merged)


def debug(title=None, **kwargs):
    if CONSOLE_DEBUG_LOGGING:
        if title:
            pprint({f"{title}": kwargs})
        else:
            pprint(kwargs)



"""
Handles storage of cache and prompt config directories, as well
as figuring out where to put them on various platforms 
"""

import os
import platformdirs
import pickle
import yaml
import random
import string

from . import errors, chat

APP_NAME = "chatblade"


def make_postfix():
    return "." + "".join(random.choices(string.ascii_letters + string.digits, k=10))


def get_cache_path(create=True):
    """
    if ~/.cache is availabe always use ~/.cache/chatblade as the cachefile
    otherwise fallback to the platform recommended location and create the directory
    e.g. ~/Library/Caches/chatblade on osx
    """
    os_cache_path = os.path.expanduser("~/.cache")
    if not os.path.exists(os_cache_path):
        os_cache_path = platformdirs.user_cache_dir(APP_NAME)
        if not os.path.exists(os_cache_path):
            os.makedirs(os_cache_path)

    cache_path = os.path.join(os_cache_path, APP_NAME)
    if create and not os.path.exists(cache_path):
        os.makedirs(cache_path)

    return cache_path


def get_session_path(session, exists=False):
    """get the path of a session file
    If exists=True, return None if the path does not exists"""
    session_path = os.path.join(get_cache_path(), f"{session}.yaml")
    if exists and not os.path.exists(session_path):
        return
    return session_path


def to_cache(messages, session):
    """cache the current messages state"""
    file_path = get_session_path(session)
    file_path_tmp = file_path + make_postfix()
    with open(file_path_tmp, "w") as f:
        yaml.dump(messages, f)
    os.replace(file_path_tmp, file_path)


def messages_from_cache(session):
    """load messages from session
    Return empty list if not exists"""
    file_path = get_session_path(session)
    if not os.path.exists(file_path):
        return []
    else:
        with open(file_path, "r") as f:
            return [chat.Message.import_yaml(m) for m in yaml.load(f, yaml.SafeLoader)]


def messages_from_cache_legacy():
    """load messages from last state or ChatbladeError if not exists"""
    file_path = get_cache_path(False)
    if not os.path.exists(file_path):
        raise errors.ChatbladeError("No last state cached from which to begin")
    else:
        with open(file_path, "rb") as f:
            return pickle.load(f)


def migrate_to_session(session):
    """save pre-session last messages to session"""
    file_path = get_cache_path(False)
    messages = messages_from_cache_legacy()
    file_path_tmp = file_path + make_postfix()
    # resolve name conflict, but keep old cache file
    # until all has gone through fine
    os.replace(file_path, file_path_tmp)
    to_cache(messages, session)
    os.unlink(file_path_tmp)


def load_prompt_file(prompt_name):
    """
    load a prompt configuration by its name
    Assumes the user created the ~/.config/chatblade/{prompt_name}
    or a file directly by path
    """
    paths_to_try = [
        prompt_name,
        os.path.expanduser(os.path.join("~/.config/chatblade", f"{prompt_name}")),
    ]
    try:
        for file_path in paths_to_try:
            if os.path.exists(file_path):
                with open(file_path, "r") as f:
                    return f.read()
        # fallback legacy
        load_prompt_config_legacy_yaml(prompt_name)
    except Exception:
        locations = "".join(["\n - " + p for p in paths_to_try])
        raise errors.ChatbladeError(
            f"no prompt {prompt_name} found in any of following locations: {locations}"
        )


def load_prompt_config_legacy_yaml(prompt_name):
    """
    LEGACY: keep right now for people that still use yaml
    load a prompt configuration by its name
    Assumes the user created the {name}.yaml in ~/.config/chatblade
    """
    path = os.path.expanduser(
        os.path.join("~/.config/chatblade", f"{prompt_name}.yaml")
    )
    try:
        with open(path, "r") as f:
            return yaml.load(f, Loader=yaml.FullLoader)["system"]
    except FileNotFoundError:
        raise errors.ChatbladeError(f"Prompt {prompt_name} not found in {path}")



import sys
import types
import os

import rich
from rich.prompt import Prompt
from rich.live import Live
from rich.text import Text

from . import printer, chat, utils, storage, errors, parser, session


def fetch_and_cache(messages, params):
    result = chat.query_chat_gpt(messages, params)
    if isinstance(result, types.GeneratorType):
        text = Text("")
        message = None
        with Live(text, refresh_per_second=4, vertical_overflow="visible") as live:
            for message in result:
                live.update(message.content)
            live.update("")
        response_msg = message
    else:
        response_msg = result
    messages.append(response_msg)
    storage.to_cache(messages, params.session or utils.scratch_session)
    return messages


def start_repl(messages, params):
    while True:
        try:
            query = Prompt.ask("[yellow]query (type 'quit' to exit)[/yellow]")
        except (EOFError, KeyboardInterrupt):
            rich.print("\n")
            exit()
        if query.lower() == "quit":
            exit()

        if not messages:
            init_msgs = (
                [storage.load_prompt_file(params.prompt_file)]
                if params.prompt_file
                else []
            )
            messages = chat.init_conversation(query, *init_msgs)
        else:
            messages.append(chat.Message("user", query))

        messages = fetch_and_cache(messages, params)
        printer.print_messages(messages[-1:], params)


def handle_input(query, params):
    utils.debug(title="cli input", query=query, params=params)

    messages = None
    if params.session:
        messages = storage.messages_from_cache(params.session)
    if messages:  # a session specified and it alredy exists
        if params.prompt_file:
            printer.warn("refusing to prepend prompt to existing session")
            exit(1)
        if query:  # continue conversation
            messages.append(chat.Message("user", query))
    else:
        if query:
            init_msgs = (
                [storage.load_prompt_file(params.prompt_file)]
                if params.prompt_file
                else []
            )
            messages = chat.init_conversation(query, *init_msgs)

    if messages:
        if params.tokens:
            token_prices = chat.get_tokens_and_costs(messages)
            printer.print_tokens(messages, token_prices, params)
        else:
            if messages[-1].role == "user":
                messages = fetch_and_cache(messages, params)
            printer.print_messages(messages, params)
    elif params.interactive:
        pass
    else:
        if params.session:
            printer.warn(
                f"session {params.session} does not exist, query is needed to initialize it"
            )
            exit(1)
        else:
            printer.warn("no query or option given. nothing to do...")
            exit(0)

    if params.interactive:
        start_repl(messages, params)


def do_session_op(sess, op, rename_to):
    if op == "list":
        print(*session.list_sessions(), sep="\n")
        return 0

    err = None
    if not sess:
        err = "session name required"
    elif op == "path" or op == "dump":
        sess_path = storage.get_session_path(sess, True)
        if sess_path:
            if op == "path":
                data = sess_path
            else:
                with open(sess_path, "r") as f:
                    data = f.read()
            print(data)
        else:
            err = "session does not exist"
    elif op == "delete":
        err = session.delete_session(sess)
    elif op == "rename":
        err = session.rename_session(sess, rename_to)
    else:
        raise ValueError(f"unknown session operation: {op}")

    if err:
        printer.warn(err)
        return 1

    return 0


def migrate_old_cache_file_if_exists():
    cache_path = storage.get_cache_path()
    if os.path.isfile(cache_path):
        try:
            storage.migrate_to_session(utils.scratch_session)
        except Exception as e:
            printer.warn(f"failed to migrate old cache file: {e}")
            return 1


def cli():
    migrate_old_cache_file_if_exists()

    query, params = parser.parse(sys.argv[1:])
    if params.session_op:
        ret = do_session_op(params.session, params.session_op, params.rename_to)
        exit(ret)
    if params.debug:
        utils.CONSOLE_DEBUG_LOGGING = True
    if params.version:
        from distutils.dist import Distribution
        dist = Distribution()
        dist.parse_config_files()
        version = dist.get_option_dict('metadata')['version'][1]
        print(f"chatblade {version}")
        exit(0)
    try:
        handle_input(query, params)
    except errors.ChatbladeError as e:
        printer.warn(e)
        exit(1)



import sys
import os
import argparse

from . import utils


def get_piped_input():
    if not sys.stdin.isatty():
        return sys.stdin.read()
    return None


def get_openai_key(options):
    if options["openai_api_key"]:
        return options["openai_api_key"]
    elif "OPENAI_API_KEY" in os.environ:
        return os.environ["OPENAI_API_KEY"]
    else:
        return None


# https://platform.openai.com/docs/models/continuous-model-upgrades
model_mappings = {
    "3.5": "gpt-3.5-turbo",
    "4": "gpt-4",
    "4t": "gpt-4-turbo",
    "4o": "gpt-4o",
    "mini": "gpt-4o-mini",
}


def get_openai_model(options):
    choice = options["chat_gpt"]
    if not choice:
        if "OPENAI_API_MODEL" in os.environ:
            choice = os.environ["OPENAI_API_MODEL"]
        else:
            choice = "mini"

    if choice in model_mappings:
        return model_mappings[choice]
    else:
        return choice


def get_theme(options):
    if options["theme"]:
        return options["theme"]
    elif "CHATBLADE_THEME" in os.environ:
        return os.environ["CHATBLADE_THEME"]
    else:
        return None


def extract_query(query):
    """The query comes from both the query and any piped input
    The user can provide either only a query, a piped input, or both
    in which case the piped input gets placed above the query
    """
    query = " ".join(query) if query else None
    piped_input = get_piped_input()
    if query and piped_input:
        return piped_input + "\n----------------\n" + query
    elif query:
        return query
    elif piped_input:
        return piped_input
    else:
        return None


def extract_options(options):
    options = vars(options)  # to map
    options["openai_api_key"] = get_openai_key(options)
    options["theme"] = get_theme(options)
    options["model"] = get_openai_model(options)
    del options["query"]
    del options["chat_gpt"]
    return utils.DotDict(options)


def valid_session(sess):
    if all(char not in sess for char in ["/", "\\", "\n"]):
        return sess
    else:
        raise argparse.ArgumentTypeError(f"invalid session name {sess}")


class RenameAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.session_op = "rename"
        try:
            namespace.rename_to = valid_session(values[0])
        except argparse.ArgumentTypeError as e:
            raise argparse.ArgumentError(self, f"target: {e}")


model_mappings_str = ", ".join(["%s (%s)" % e for e in model_mappings.items()])


def parse(args):
    parser = argparse.ArgumentParser(
        "Chatblade",
        description="a CLI Swiss Army Knife for ChatGPT",
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=35),
    )
    parser.add_argument("query", type=str, nargs="*", help="Query to send to chat GPT")

    parser.add_argument(
        "--openai-api-key",
        metavar="key",
        type=str,
        help="the OpenAI API key can also be set as env variable OPENAI_API_KEY",
    )
    parser.add_argument(
        "--openai-base-url",
        metavar="key",
        type=str,
        help="A custom url to use the openAI against a local or custom model, eg ollama",
    )
    parser.add_argument(
        "--temperature",
        metavar="t",
        type=float,
        help="temperature (openai setting)",
        default=0.0,
    )
    parser.add_argument(
        "-c",
        "--chat-gpt",
        help=f"""chat GPT model use either the fully qualified model name, or
        {model_mappings_str}. Can also be set via env variable OPENAI_API_MODEL
        """,
        type=str,
    )
    parser.add_argument(
        "-i",
        "--interactive",
        help="start an interactive chat session. This will implicitly continue the conversation",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--stream",
        help="Stream the incoming text to the terminal",
        action="store_true",
    )
    parser.add_argument(
        "-t",
        "--tokens",
        help="display what *would* be sent, how many tokens, and estimated costs",
        action="store_true",
    )
    parser.add_argument(
        "--version",
        help="display the chatblade version",
        action="store_true",
    )
    parser.add_argument(
        "-p",
        "--prompt-file",
        metavar="name",
        type=str,
        help="prompt name - will load the prompt with that name at ~/.config/chatblade/name or a path to a file",
    )

    display_opts = parser.add_argument_group("result formatting options")
    display_opts.add_argument(
        "-e",
        "--extract",
        help="extract content from response if possible (either json or code block)",
        action="store_true",
    )
    display_opts.add_argument(
        "-r",
        "--raw",
        help="print session as pure text, don't pretty print or format",
        action="store_true",
    )
    display_opts.add_argument(
        "-n",
        "--no-format",
        help="do not add pretty print formatting to output",
        action="store_true",
    )
    display_opts.add_argument(
        "-o",
        "--only",
        help="Only display the response, omit query",
        action="store_true",
    )
    display_opts.add_argument(
        "--theme",
        metavar="theme",
        type=str,
        help="Set the theme for syntax highlighting see https://pygments.org/styles/, can also be set with CHATBLADE_THEME",
    )

    session_opts = parser.add_argument_group("session options")
    session_opts.add_argument(
        "-l",
        "--last",
        dest="session",
        action="store_const",
        const=utils.scratch_session,
        help=f"alias for '-S {utils.scratch_session}', the default session if none is specified",
    )
    session_opts.add_argument(
        "-S",
        "--session",
        metavar="sess",
        type=valid_session,
        help="""initiate or continue named session""",
    )
    session_opts.add_argument(
        "--session-list",
        dest="session_op",
        action="store_const",
        const="list",
        help="list sessions",
    )
    session_opts.add_argument(
        "--session-path",
        dest="session_op",
        action="store_const",
        const="path",
        help="show path to session file",
    )
    session_opts.add_argument(
        "--session-dump",
        dest="session_op",
        action="store_const",
        const="dump",
        help="dump session to stdout",
    )
    session_opts.add_argument(
        "--session-delete",
        dest="session_op",
        action="store_const",
        const="delete",
        help="delete session",
    )
    session_opts.add_argument(
        "--session-rename",
        metavar="newsess",
        action=RenameAction,
        nargs=1,
        help="rename session",
    )

    # --- debug
    parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)

    options = parser.parse_args(args)
    return extract_query(options.query), extract_options(options)
